#!/usr/bin/env python3

import numpy
import struct
from Crypto.Util.number import bytes_to_long, long_to_bytes

def parse_keystore(filename):
    keys = []
    with open(filename, "rb") as f:
        while True:
            if f.read(8) != b"KEY".ljust(8, b"\x00"):
                break
            flags, length = struct.unpack("<LL", f.read(8))
            key_data = f.read(length)
            keys.append((flags, key_data))
    return keys

def main():
    ks = parse_keystore("keystore_init")
    flags, aes_key = ks[1]
    if flags != 3:
        print("AES Key already decrypted: %s" % aes_key)
        return
    n = bytes_to_long(ks[0][1][:0x80])
    print("n: 0x%x" % n)

    power_consumption = numpy.load("virtual_phone/power_consumption.npz")
    samples = list(power_consumption.items())[0][1]

    start = 700000
    end = 1490000 + 10000

    private_exp = 0
    last = None
    count = 0
    bits = 0
    for i in range(start, end):
        x = samples[i]
        if x < -0.15:
            if not last or ((i - last) < 500):
                count += 1
            else:
                private_exp |= (int(count >= 4) << bits)
                bits += 1
                if bits == 1024:
                    print("got enough bits at %d" % i)
                    break
                count = 1
            last = i
    assert bits == 1024
    dec = long_to_bytes(pow(bytes_to_long(aes_key), private_exp, n)).rjust(0x80, b"\x00")
    assert dec[0] == 0 and dec[95] == 0 and dec[1] == 2
    print("RSA private exponent: 0x%x" % private_exp)
    print("AES Key: %s" % dec[-32:])

if __name__ == "__main__":
    main()
