#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <assert.h>
#include <string.h>
#include <time.h>
#include "Sharlotte256/Sharlotte256.h"

static uint32_t B = 0;

unsigned rol1(uint8_t a) {
    return (a << 1) | (a >> (7));
}

static uint8_t secure_device(uint8_t a, uint8_t b, uint8_t op)
{
    if(B & 4)
    {
        a = rol1(a);
    }
    if(B & 8)
    {
        b = rol1(b);
    }

    uint8_t sel = op ^ (B & 3);
    switch(sel)
    {
        case 0:
            return a&b;
        case 1:
            return a|b;
        case 2:
            return a^b;
        case 3:
            return a+b;
    }
    return 0;
}

static uint8_t init()
{
    uint8_t r;
    r = secure_device(0x46,0x92,0);
    r = secure_device(0xdf,r,2);
    r = secure_device(0x3e,r,0);
    r = secure_device(0x3a,r,3);
    r = secure_device(0x36,r,2);
    r = secure_device(0x8e,r,2);
    r = secure_device(0xc9,r,3);
    r = secure_device(0xe7,r,1);
    r = secure_device(0x29,r,2);
    r = secure_device(0xc2,r,2);
    r = secure_device(0x79,r,0);
    r = secure_device(0x2a,r,2);
    r = secure_device(0x4c,r,3);
    r = secure_device(0xde,r,0);
    r = secure_device(0x88,r,0);
    r = secure_device(0x8b,r,2);
    r = secure_device(0x97,r,3);
    r = secure_device(0x6a,r,2);
    r = secure_device(0x60,r,1);
    r = secure_device(0x0f,r,0);
    r = secure_device(0x5b,r,3);
    r = secure_device(0xd0,r,2);
    r = secure_device(0xa9,r,1);
    r = secure_device(0xe3,r,3);
    r = secure_device(0xd0,r,1);
    r = secure_device(0x27,r,0);
    r = secure_device(0x90,r,0);
    r = secure_device(0x3b,r,1);
    r = secure_device(0x66,r,2);
    r = secure_device(0xe2,r,0);
    r = secure_device(0x24,r,3);
    r = secure_device(0xee,r,1);
    r = secure_device(0xf2,r,3);
    return r;
}


static uint8_t step1()
{
    uint8_t r;
    r = secure_device(0x35,0x27,3);
    r = secure_device(0x7e,r,3);
    r = secure_device(0x66,r,2);
    r = secure_device(0x8,r,1);
    r = secure_device(0x13,r,0);
    r = secure_device(0x1f,r,1);
    r = secure_device(0xa,r,2);
    r = secure_device(0xd3,r,0);
    r = secure_device(0xc6,r,3);
    return r;
};

static uint8_t step2()
{
    uint8_t r;
    r = secure_device(0xde,0xab,0);
    r = secure_device(0x67,r,3);
    r = secure_device(0x2a,r,2);
    r = secure_device(0x6d,r,1);
    r = secure_device(0x4a,r,3);
    r = secure_device(0xe7,r,0);
    r = secure_device(0x1c,r,1);
    r = secure_device(0x35,r,0);
    r = secure_device(0xde,r,3);
    r = secure_device(0xf7,r,0);
    r = secure_device(0xda,r,2);
    return r;
}

static uint8_t step3()
{
    uint8_t r = secure_device(0x14,0x23,3);
    r = secure_device(0x72,r,0);
    r = secure_device(0x48,r,3);
    r = secure_device(0x53,r,1);
    r = secure_device(0xa7,r,0);
    r = secure_device(0x5f,r,1);
    r = secure_device(0x3,r,3);
    r = secure_device(0xb7,r,3);
    r = secure_device(0x73,r,1);
    r = secure_device(0x37,r,3);
    r = secure_device(0xc5,r,2);
    r = secure_device(0xa4,r,1);
    r = secure_device(0x30,r,0);
    r = secure_device(0xdd,r,2);
    return r;
}

static uint8_t step4()
{
    uint8_t r;
    r = secure_device(0xb0,0x42,2);
    r = secure_device(0xbc,r,2);
    r = secure_device(0xfc,r,2);
    r = secure_device(0x54,r,3);
    r = secure_device(0x30,r,2);
    r = secure_device(0x97,r,1);
    r = secure_device(0xe8,r,2);
    r = secure_device(0xd6,r,0);
    r = secure_device(0x26,r,0);
    r = secure_device(0xeb,r,0);
    r = secure_device(0x68,r,1);
    r = secure_device(0x26,r,0);
    r = secure_device(0x9,r,3);
    r = secure_device(0x2a,r,2);
    r = secure_device(0xa9,r,3);
    return r;
}

static uint8_t step5()
{
    uint8_t r;
    r = secure_device(0xff,0x12,0);
    r = secure_device(0xfd,r,1);
    r = secure_device(0xe5,r,1);
    r = secure_device(0x26,r,3);
    r = secure_device(0x85,r,3);
    r = secure_device(0x63,r,1);
    r = secure_device(0x93,r,3);
    r = secure_device(0xba,r,2);
    r = secure_device(0x97,r,0);
    r = secure_device(0xab,r,1);
    r = secure_device(0x6e,r,3);
    r = secure_device(0xfd,r,0);
    r = secure_device(0x4c,r,3);
    r = secure_device(0x50,r,0);
    r = secure_device(0xa,r,2);
    r = secure_device(0xfc,r,3);
    r = secure_device(0xe3,r,2);
    r = secure_device(0xa6,r,3);
    r = secure_device(0x64,r,2);
    r = secure_device(0x8e,r,3);
    r = secure_device(0xc1,r,1);
    return r;
}

static uint8_t step6()
{
    uint8_t r;
    r = secure_device(0x90,0x77,1);
    r = secure_device(0x8e,r,0);
    r = secure_device(0xbd,r,2);
    r = secure_device(0x39,r,2);
    r = secure_device(0x4c,r,2);
    r = secure_device(0xc5,r,2);
    r = secure_device(0xb6,r,3);
    r = secure_device(0x93,r,1);
    r = secure_device(0x9f,r,3);
    r = secure_device(0xd6,r,3);
    r = secure_device(0x6e,r,2);
    r = secure_device(0x39,r,3);
    r = secure_device(0x40,r,1);
    r = secure_device(0x14,r,2);
    r = secure_device(0xe6,r,3);
    return r;
}

static uint8_t step7()
{
    uint8_t r;
    r = secure_device(0xf,0xab,3);
    r = secure_device(0xa2,r,1);
    r = secure_device(0x7c,r,0);
    r = secure_device(0x34,r,1);
    r = secure_device(0x14,r,1);
    r = secure_device(0xe7,r,0);
    r = secure_device(0xb9,r,0);
    r = secure_device(0xf1,r,2);
    r = secure_device(0xd5,r,1);
    r = secure_device(0x4e,r,2);
    r = secure_device(0xe,r,2);
    r = secure_device(0x6,r,0);
    r = secure_device(0x7d,r,2);
    r = secure_device(0x87,r,3);
    r = secure_device(0xbc,r,0);
    r = secure_device(0xd4,r,3);
    r = secure_device(0x8a,r,1);
    r = secure_device(0xe7,r,3);
    r = secure_device(0x9e,r,1);
    r = secure_device(0x58,r,0);
    r = secure_device(0x24,r,2);
    r = secure_device(0x44,r,3);
    r = secure_device(0xc9,r,1);
    r = secure_device(0xd4,r,1);
    r = secure_device(0x1d,r,3);
    r = secure_device(0xcd,r,0);
    r = secure_device(0xde,r,1);
    r = secure_device(0x54,r,0);
    r = secure_device(0x5e,r,2);
    r = secure_device(0x46,r,1);
    r = secure_device(0x21,r,0);
    r = secure_device(0xff,r,1);
    r = secure_device(0x51,r,0);
    r = secure_device(0x78,r,1);
    r = secure_device(0x2f,r,3);
    r = secure_device(0xed,r,2);
    r = secure_device(0x4b,r,3);
    r = secure_device(0x4d,r,2);
    return r;
}

static uint8_t step8()
{
    uint8_t r;
    r = secure_device(0x88,0x74,0);
    r = secure_device(0x48,r,2);
    r = secure_device(0x11,r,2);
    r = secure_device(0x76,r,0);
    r = secure_device(0x2b,r,3);
    r = secure_device(0xf8,r,2);
    return r;
}

int test_init(uint8_t b_state, uint8_t expected)
{
    B = b_state;
    uint8_t res = init();
    return res == expected;
}

int main(int argc, const char * argv[]) {
    assert(test_init(0xF, 0xa1));
    assert(test_init(0x0, 0xe0));
    
    uint8_t key[8] = {0};
    unsigned char md[32] = {0};
    uint8_t target[32] = {0x00, 0xc8, 0xbb, 0x35, 0xd4, 0x4d, 0xcb, 0xb2, 0x71, 0x2a, 0x11, 0x79,
                          0x9d, 0x8e, 0x13, 0x16, 0x04, 0x5d, 0x64, 0x40, 0x4f, 0x33, 0x7f, 0x4f,
                          0xf6, 0x53, 0xc2, 0x76, 0x07, 0xf4, 0x36, 0xea};

    for(uint32_t x=0; x < 0xffffffff; x++)
    {
        B = x;
        key[0] = step1();
        B >>= 4;
        key[1] = step2();
        B >>= 4;
        key[2] = step3();
        B >>= 4;
        key[3] = step4();
        B >>= 4;
        key[4] = step5();
        B >>= 4;
        key[5] = step6();
        B >>= 4;
        key[6] = step7();
        B >>= 4;
        key[7] = step8();

        Sharlotte256(key, sizeof(key), md);

        if(!memcmp(md, target, sizeof(target)))
        {
            printf("x=0x%08x\n", x);
            printf("key= %02X%02X%02X%02X%02X%02X%02X%02X\n", key[0], key[1], key[2], key[3], key[4], key[5], key[6], key[7]);
            break;
        }
    }

    return 0;
}
