#!/usr/bin/env python

import struct
import subprocess

DW_OP_addr	= 0x03
DW_OP_deref	= 0x06
DW_OP_const1u	=0x08
DW_OP_const1s	=0x09
DW_OP_const2u	=0x0a
DW_OP_const2s	=0x0b
DW_OP_const4u	=0x0c
DW_OP_const4s	=0x0d
DW_OP_const8u	=0x0e
DW_OP_const8s	=0x0f
DW_OP_constu	=0x10
DW_OP_consts	=0x11
DW_OP_dup	= 0x12
DW_OP_drop	= 0x13
DW_OP_over	= 0x14
DW_OP_pick	= 0x15
DW_OP_swap	= 0x16
DW_OP_rot	= 0x17
DW_OP_xderef	=0x18
DW_OP_abs	= 0x19
DW_OP_and	= 0x1a
DW_OP_div	= 0x1b
DW_OP_minus	= 0x1c
DW_OP_mod	= 0x1d
DW_OP_mul	= 0x1e
DW_OP_neg	= 0x1f
DW_OP_not	= 0x20
DW_OP_or	= 0x21
DW_OP_plus	= 0x22
DW_OP_plus_uconst	= 0x23
DW_OP_shl	= 0x24
DW_OP_shr	= 0x25
DW_OP_shra	= 0x26
DW_OP_xor	= 0x27
DW_OP_skip	= 0x2f
DW_OP_bra	= 0x28
DW_OP_eq	= 0x29
DW_OP_ge	= 0x2a
DW_OP_gt	= 0x2b
DW_OP_le	= 0x2c
DW_OP_lt	= 0x2d
DW_OP_ne	= 0x2e
DW_OP_lit0	= 0x30
DW_OP_lit1	= 0x31
DW_OP_lit2	= 0x32
DW_OP_lit3	= 0x33
DW_OP_lit4	= 0x34
DW_OP_lit5	= 0x35
DW_OP_lit6	= 0x36
DW_OP_lit7	= 0x37
DW_OP_lit8	= 0x38
DW_OP_lit9	= 0x39
DW_OP_lit10	= 0x3a
DW_OP_lit11	= 0x3b
DW_OP_lit12	= 0x3c
DW_OP_lit13	= 0x3d
DW_OP_lit14	= 0x3e
DW_OP_lit15	= 0x3f
DW_OP_lit16	= 0x40
DW_OP_lit17	= 0x41
DW_OP_lit18	= 0x42
DW_OP_lit19	= 0x43
DW_OP_lit20	= 0x44
DW_OP_lit21	= 0x45
DW_OP_lit22	= 0x46
DW_OP_lit23	= 0x47
DW_OP_lit24	= 0x48
DW_OP_lit25	= 0x49
DW_OP_lit26	= 0x4a
DW_OP_lit27	= 0x4b
DW_OP_lit28	= 0x4c
DW_OP_lit29	= 0x4d
DW_OP_lit30	= 0x4e
DW_OP_lit31	= 0x4f
DW_OP_reg0	= 0x50
DW_OP_reg1	= 0x51
DW_OP_reg2	= 0x52
DW_OP_reg3	= 0x53
DW_OP_reg4	= 0x54
DW_OP_reg5	= 0x55
DW_OP_reg6	= 0x56
DW_OP_reg7	= 0x57
DW_OP_reg8	= 0x58
DW_OP_reg9	= 0x59
DW_OP_reg10	= 0x5a
DW_OP_reg11	= 0x5b
DW_OP_reg12	= 0x5c
DW_OP_reg13	= 0x5d
DW_OP_reg14	= 0x5e
DW_OP_reg15	= 0x5f
DW_OP_reg16	= 0x60
DW_OP_reg17	= 0x61
DW_OP_reg18	= 0x62
DW_OP_reg19	= 0x63
DW_OP_reg20	= 0x64
DW_OP_reg21	= 0x65
DW_OP_reg22	= 0x66
DW_OP_reg23	= 0x67
DW_OP_reg24	= 0x68
DW_OP_reg25	= 0x69
DW_OP_reg26	= 0x6a
DW_OP_reg27	= 0x6b
DW_OP_reg28	= 0x6c
DW_OP_reg29	= 0x6d
DW_OP_reg30	= 0x6e
DW_OP_reg31	= 0x6f
DW_OP_breg0	= 0x70
DW_OP_breg1	= 0x71
DW_OP_breg2	= 0x72
DW_OP_breg3	= 0x73
DW_OP_breg4	= 0x74
DW_OP_breg5	= 0x75
DW_OP_breg6	= 0x76
DW_OP_breg7	= 0x77
DW_OP_breg8	= 0x78
DW_OP_breg9	= 0x79
DW_OP_breg10	= 0x7a
DW_OP_breg11	= 0x7b
DW_OP_breg12	= 0x7c
DW_OP_breg13	= 0x7d
DW_OP_breg14	= 0x7e
DW_OP_breg15	= 0x7f
DW_OP_breg16	= 0x80
DW_OP_breg17	= 0x81
DW_OP_breg18	= 0x82
DW_OP_breg19	= 0x83
DW_OP_breg20	= 0x84
DW_OP_breg21	= 0x85
DW_OP_breg22	= 0x86
DW_OP_breg23	= 0x87
DW_OP_breg24	= 0x88
DW_OP_breg25	= 0x89
DW_OP_breg26	= 0x8a
DW_OP_breg27	= 0x8b
DW_OP_breg28	= 0x8c
DW_OP_breg29	= 0x8d
DW_OP_breg30	= 0x8e
DW_OP_breg31	= 0x8f
DW_OP_regx	= 0x90
DW_OP_fbreg	= 0x91
DW_OP_bregx	= 0x92
DW_OP_piece	= 0x93
DW_OP_deref_size	= 0x94
DW_OP_xderef_size	= 0x95
DW_OP_nop	= 0x96
DW_OP_push_object_address	= 0x97
DW_OP_call2	= 0x98
DW_OP_call4	= 0x99
DW_OP_call_ref	= 0x9a
DW_OP_form_tls_address	= 0x9b
DW_OP_call_frame_cfa	= 0x9c
DW_OP_bit_piece	= 0x9d
DW_OP_lo_user	= 0xe0
DW_OP_hi_user	= 0xff

class CodeGenerator(object):
    def __init__(self):
        self.stack_elt = 0
        self.addr = None
        self.code = r"""
#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <stdlib.h>
#include <assert.h>
#include "dwarf_data.h"
#include "hexdump.h"

typedef struct {
    char pad[0xA8];
    char** ptr_to_input;
} struct1;

unsigned long* func1(struct1* reg31, unsigned initial, uint32_t* static_data);
unsigned long func2(unsigned long input);

uint32_t __func2(uint32_t input)
{
//    assert((input == 0x17246549) || (input == 0x84653217));
    if(input == 0x17246549) return 0x818f694a;
    else  return  0x60bf080f;
}


int main(int argc, char** argv)
{
    char* flag_32_A = "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA";
    char* flag = flag_32_A;

    uint64_t target[5] = {0x65850b36e76aaed5, 0xd9c69b74a86ec613, 0xdc7564f1612e5347, 0x658302a68e8e1c24, 0x0};
    //flag = (char*) target;

    if(argc == 2 && strlen(argv[1])==32)
    {
        flag = argv[1];
    }

    printf("func2 0x%x => 0x%x\n", 0x17246549, __func2(0x17246549));
    printf("func2 0x%x => 0x%x\n", 0x84653217, __func2(0x84653217));

    char* tab[2] = {0x0, flag};
    struct1 s = {0};
    s.ptr_to_input = tab;
    unsigned long* res = func1(&s, 0x0, (uint32_t*) dwarf_data_bin);
    return 0;
}

unsigned long* func1(struct1* reg31, unsigned initial, uint32_t* static_data)
{
    unsigned long stack[64] = {0};
    unsigned long t,t1,t2,t3, first, second;
    int stack_elt;
    void* x = NULL;
    stack[0] = initial;
    stack_elt = 1;
"""
        self.stack_elt = 1

    def addline(self, l):
        if self.addr:
            self.code += "\nlabel_%04x:\n\n" % self.addr
            if self.addr == 0x163:
                self.code += r"""printf("loop stack_elt=%d\n", stack_elt);"""
            elif self.addr == 0x2d:
                self.code += r"""
        printf("got 0x%lx expected 0x%lx\n", stack[24], 0xe4664dd59de6a860);
            if(stack[24] == 0xe4664dd59de6a860) printf("OK !\n");
            """
            self.addr = None
        if not l.endswith(";"):
            l += ";"
        self.code += "    %s\n" % l

    def done_func1(self):
        self.addline("""func1_end:
        /*x = calloc(1, sizeof(stack));
        memcpy(x, stack, sizeof(stack));
        return x;*/
        return NULL;}""")
    
    def start_func2(self):
        self.addline("""
unsigned long func2(unsigned long input)
{
    unsigned long stack[64] = {0};
    unsigned long t,t1,t2,t3, first, second;
    int stack_elt;
    stack[0] = input;
    stack_elt = 1;
    """)
        self.stack_elt = 1

    def done(self):
        if False:
            self.addline(r"""
                return stack[stack_elt-1];
            }""")
        sourcefile = "3_dwarf_decompile.c"
        with open(sourcefile,"w") as f:
            f.write(self.code)
        
        if subprocess.call(["clang", "-Ofast", "-flto", sourcefile]) == 0:
            subprocess.call(["./a.out"])

    def push_result(self, res):
        self.addline("stack[stack_elt] = (unsigned long) (%s);stack_elt++;" % res)
        self.stack_elt += 1

    def DW_OP_reg31(self, value):
        self.push_result("reg31")

    def DW_OP_swap(self, value):
        self.addline("t = stack[stack_elt - 1];")
        self.addline("stack[stack_elt - 1] = stack[stack_elt - 2];")
        self.addline("stack[stack_elt - 2] = t;")

    def DW_OP_rot(self, value):
        self.addline("""t1 = stack[stack_elt - 1];
    t2 = stack[stack_elt - 2];
    t3 = stack[stack_elt - 3];
    stack[stack_elt - 1] = t2;
    stack[stack_elt - 2] = t3;
    stack[stack_elt - 3] = t1;""")

    def DW_OP_dup(self, value):
        self.push_result("stack[stack_elt - 1]")

    def DW_OP_drop(self, value):
        self.addline("stack_elt -= 1;")
        self.stack_elt -= 1

    def DW_OP_pick(self, value):
        self.push_result("stack[stack_elt - 1 - 0x%x]" % value)

    def operation(self, op):
        self.addline("""stack_elt -= 2;
    second = stack[stack_elt];
    first = stack[stack_elt + 1];""")
        self.stack_elt -= 2
        self.push_result("(second %s first)" % op);

    def DW_OP_and(self, value):
        self.operation("&")
    def DW_OP_or(self, value):
        self.operation("|")
    def DW_OP_xor(self, value):
        self.operation("^")
    def DW_OP_mul(self, value):
        self.operation("*")
    def DW_OP_plus(self, value):
        self.operation("+")
    def DW_OP_minus(self, value):
        self.operation("-")
    def DW_OP_shr(self, value):
        self.operation(">>")
    def DW_OP_shl(self, value):
        self.operation("<<")

    def DW_OP_const1u(self, value):
        self.push_result("0x%x" % value)
    def DW_OP_const4u(self, value):
        self.push_result("0x%x" % value)

    def DW_OP_const8u(self, value):
        if value >= 0x400648 and value < 0x400700:
            data_base = 0x400648
            self.push_result("(unsigned long) (&((uint8_t*)static_data)[0x%x])" % (value-data_base))
        else:
            self.push_result("0x%x" % value)

    def DW_OP_lit0(self, value):
        self.push_result("0x0")
    def DW_OP_lit1(self, value):
        self.push_result("0x1")
    def DW_OP_lit2(self, value):
        self.push_result("0x2")
    def DW_OP_lit4(self, value):
        self.push_result("0x4")
    def DW_OP_lit6(self, value):
        self.push_result("0x6")
    def DW_OP_lit15(self, value):
        self.push_result("0xf")

    def DW_OP_deref_size(self, value):
        addr = self.addr
        sz = value*8
        self.addline("stack_elt -= 1")
        self.stack_elt -= 1
        self.addline("t = stack[stack_elt]")
        self.addline("t = *((uint%d_t*) t)" % sz)
        self.push_result("t")

    def DW_OP_deref(self, value):
        addr = self.addr
        self.addline("stack_elt -= 1")
        self.stack_elt -= 1
        self.addline("t = stack[stack_elt]")
        self.push_result("*((uintptr_t*) t)")

    def DW_OP_bra(self, value):
        self.addline("""stack_elt -= 1;
    if (stack[stack_elt] != 0)
        goto label_%04x""" % value)
        self.stack_elt -= 1

    def DW_OP_skip(self, value):
        if value == 0x8069:
            self.addline(r"""printf("That's the correct flag :)\n");return NULL;""")
        elif value == 0x8075:
            self.addline(r"""printf("Not good\n");return NULL;""")
        else:
            if value == 0x85d:
                self.addline(r"""stack[stack_elt-1] = __func2(stack[stack_elt-1]);""")
            elif value == 0x0398:
                #0xbdedea50f5a2b2bd
                #self.addline(r"""printf("return 0x85d stack_elt=%d\n", stack_elt);""")
                pass#self.addline("return;")
            else:
                self.addline("goto label_%04x" % value)


"""
A*32
=>
0xa91c162e2b2d1ffd 0x56d94e844b176781	0x252afb8890764425	0xe4664dd59de6a86
"""
def main():
    op_names = {}
    for x in globals():
        if x.startswith("DW_OP"):
            op_names[globals()[x]] = x

    symbols = {0x04030b8: "Not good",
                0x403098: "That's the correct flag :)"}

    codegen = CodeGenerator()
    dwarf_code_offset = 0x258
    dwarf_data_offset = 0x648
    f = open("root/safe_01/decrypted_file", "rb")
    f.seek(dwarf_data_offset)
    dwarf_data = f.read(1132)
    with open("dwarf_data.bin", "wb") as f2:
        f2.write(dwarf_data)
    dwarf_data_c = subprocess.check_output(["xxd", "-i", "dwarf_data.bin"])
    with open("dwarf_data.h", "wb") as f2:
        f2.write(dwarf_data_c)
    f.seek(dwarf_code_offset)
    ins = set()
    for i in range(5000):
        off = f.tell() - dwarf_code_offset
        op = ord(f.read(1))
        if op == DW_OP_nop:
            print("%x" % off)
            codegen.done_func1()
            break
            codegen.start_func2()
            f.seek(dwarf_code_offset + 0x85d, 0)
            continue#break
        if op == 0:
            break
        if op not in op_names:
            print("Unknown opcode 0x%x" % op)
            return
        desc = name = op_names[op]
        ins.add(name)
        param_size = 0
        value = None
        if name.startswith("DW_OP_const"):
            param_size = int(name.replace("DW_OP_const","").replace("u",""))
        elif name in ["DW_OP_deref_size", "DW_OP_pick"]:
            param_size = 1
        elif name in ["DW_OP_bra", "DW_OP_skip"]:
            param_size = 2
        elif name == "DW_OP_addr":
            param_size = 8

        if param_size:
            param = f.read(param_size)
            if param_size == 1:
                value = ord(param)
            desc = "%s %s" % (name, param[::-1].hex())

        if name in ["DW_OP_bra", "DW_OP_skip"]:
            skip = struct.unpack("<h", param)[0]
            desc = "%s %s 0x%x => 0x%04x" % (name, param.hex(), skip, off+3+skip)
            value = off+3+skip
        elif op == DW_OP_const4u:
            value = struct.unpack("<L", param)[0]
            desc = "DW_OP_const4u 0x%x (%s)" % (value, symbols.get(value, ""))
        elif op == DW_OP_const8u:
            value = struct.unpack("<Q", param)[0]
            desc = "DW_OP_const8u 0x%x (%s)" % (value, symbols.get(value, ""))

        if True:
            print("0x%04x:      %s" % (off, desc))
        if hasattr(codegen, name):
            codegen.addr = off
            getattr(codegen, name)(value)
        else:
            raise

    print("%d unique instructions" % len(ins))
    for x in ins:
        if not hasattr(codegen, x):
            print(x)

    codegen.done()

if __name__ == "__main__":
    main()