#!/usr/bin/env python3
import struct
import subprocess

def unobfu_str(s):
    return "".join(map(chr, [~ord(x) & 0xff for x in s]))

def strings_gdb():
    s1 = "d08f8d909cd08c9a9399d09c929b9396919aff".decode("hex")
    s2 = "d78f8cdfd29a99df839a988d9a8fdfd28edfd88e9a928ad28c868c8b9a92d1d5dfd2a48ca2df838e9a928ad28c868c8b9a92d1d5dfd2a498a29b9ddfd8d6dfd9d9df9a87968bdfcbcdff00".decode("hex")
    s3 = "\xD2\x98\x9b\x9d"
    print(unobfu_str(s1))
    print(unobfu_str(s2)[:-2])
    print(unobfu_str(s3))
    stuff = [0x612e7270, 0x6766722e, 0x666e632e, 0x2e76662e, 0x76706e73, 0x66407279, 0x70766766, 0x7465622e]
    print("".join([struct.pack("<L", x) for x in stuff]).decode("rot13"))

def disassemble_ins(pc, ins):
    a = (ins >> 20) & 0b1111
    b = (ins >> 18) & 0b0011
    c = (ins >> 14) & 0b1111
    d = (ins >> 10) & 0b1111
    dd = (ins >> 4)  & 0b111111
    e = (ins >> 0)  & 0b1111
    imm = ins & 0x3fff #15bits

    disas = ""
    if a == 0xA:
        disas = "exit"
        cc = "/*dump_regs(regs);*/return r0;"
    elif a == 0xC:
        disas = "store_magic_dword at 0x100000[r%d]" % c
        cc = "store_magic(r%d);" % c
    elif a == 0xD:
        disas = "clear_counter 0x9010000"
        cc = "clear_tsc();"
    elif a == 0xe:
        disas = "branch_if_counter_gt_5 0x9010000 0x%x" % (imm)
        cc = "//if(get_tsc()>5) goto label_%04x;" % imm
    elif b == 0x0:
        if a == 0:
            disas = "mov r%d, r%d" % (c,d)
            cc = "r%d = r%d;" % (c,d)
        elif a == 1:
            disas = "dec r%d" % (c)
            cc = "r%d--;" % (c)
        elif a == 2:
            disas = "add r%d, r%d" % (c,d)
            cc = "r%d += r%d;" % (c,d)
        elif a == 3:
            disas = "sub r%d, r%d" % (c,d)
            cc = "r%d -= r%d;" % (c,d)
        elif a == 6:
            disas = "xor r%d, r%d" % (c,d)
            cc = "r%d ^= r%d;" % (c,d)
        elif a == 0xb:
            disas = "bswap16 r%d, r%d" % (c,d)
            cc = "r%d = (r%d  >> 8) | ((r%d & 0xff) << 8);" % (c,c, c)
    elif b == 0x1:
        disas = "mov r%d, [r%d]" % (c,d)
        cc = "r%d = read_dword_at(r%d);" % (c,d)
    elif b == 0x2:
        disas = "mov [r%d], r%d" % (c,d)
        cc = "write_dword_at(r%d,r%d);" % (c,d)
    elif b == 0x3:
        if a == 0:
            disas = "mov r%d, 0x%x" % (c, imm)
            cc = "r%d = 0x%x;" % (c,imm)
        elif a == 1:
            raise
        elif a == 2:
            disas = "add r%d, 0x%x" % (c, imm)
            cc = "r%d += 0x%x;" % (c,imm)
        elif a == 3:
            disas = "sub r%d, 0x%x" % (c, imm)
            cc = "r%d -= 0x%x;" % (c,imm)
        elif a == 4:
            disas = "shl r%d, 0x%x" % (c, imm)
            cc = "r%d <<= 0x%x;" % (c,imm)
        elif a == 5:
            disas = "shr r%d, 0x%x" % (c, imm)
            cc = "r%d >>= 0x%x;" % (c,imm)
        elif a == 7:
            disas = "and r%d 0x%x" % (c, imm)
            cc = "r%d &=  0x%x;" % (c,imm)
        elif a == 8:
            disas = "branch 0x%x" % (imm)
            assert (imm % 3) == 0
            cc = "goto label_%04x;" % imm 
        elif a == 9:
            disas = "bnz r%d 0x%x" % (c,imm)
            cc = "if(r%d) goto label_%04x;" % (c,imm)
            assert (imm % 3) == 0
    disas = "0x%04x: %06x %x %x %x %x %x   => %s" % (pc, ins, a, b, c, d, e, disas)
    assert cc
    pr = ""
    if pc == 0x0186:
        pr += "printf(\"check\\n\");"
        pr += "hexdump(&memory_bin[0x100000], 0x20);"
        pr += "hexdump(&memory_bin[0x100020], 0x20);"
    return disas, cc +pr

MEMORY_DUMP_FILE = "4_vm_memory.bin"

def main_generate_c():
    memory = open(MEMORY_DUMP_FILE, "rb").read()

    vm_memory_c = subprocess.check_output(["xxd", "-i", "4_vm_memory.bin"])
    with open("4_vm_memory.h", "wb") as f2:
        f2.write(vm_memory_c)

    c_code = r"""
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <assert.h>
#include "4_vm_memory.h"
#include "hexdump.h"

#define memory_bin __4_vm_memory_bin

uint32_t do_stuff();
static uint32_t tsc = 0;

static int debug = 0;
static inline uint32_t read_dword_at(uint32_t addr)
{
    if(debug)
    {
        printf("MEM: read at 0x%x\n", addr);
        //assert(addr < __4_vm_memory_bin_len);
    }
    return *((uint32_t*) (&__4_vm_memory_bin[addr]));
}
static inline void write_dword_at(uint32_t addr, uint32_t value)
{
    if(debug)
    {
        printf("MEM:write at 0x%x 0x%x\n", addr, value);
        //assert(addr < __4_vm_memory_bin_len);
    }
    *((uint32_t*) (&__4_vm_memory_bin[addr])) = value;
}
static inline void clear_tsc()
{
    tsc = 0;
}
static inline uint32_t get_tsc()
{
    return 0;
}
static inline void store_magic(uint32_t offset)
{
    uint32_t values[8] = {0x612e7270, 0x6766722e, 0x666e632e, 0x2e76662e,
        0x76706e73, 0x66407279, 0x70766766,  0x7465622e};
    offset = (offset>>2) & 7;
    assert(offset < 8);
}

void dump_regs(uint32_t* regs)
{
    for(int i=0; i < 16; i++)
    {
        printf("r%d = 0x%08x\n", i ,regs[i]);
    }
}
//https://gist.github.com/clzola/b7f1d9dfacab7a253afce7ca3d5aeb23
unsigned char* hex2bin(const char* hexstr, size_t* size)
{
    size_t hexstrLen = strlen(hexstr);
    size_t bytesLen = hexstrLen / 2;

    unsigned char* bytes = (unsigned char*) malloc(bytesLen);

    int count = 0;
    const char* pos = hexstr;

    for(count = 0; count < bytesLen; count++) {
        sscanf(pos, "%2hhx", &bytes[count]);
        pos += 2;
    }

    if( size != NULL )
        *size = bytesLen;

    return bytes;
}
int main(int argc, char** argv)
{
    uint8_t testcase[32] = {0};
    memset(testcase, 0xAA, 32);
    const uint8_t check[32] = {
        0x41, 0x73, 0x93, 0x33, 0xE4, 0x09, 0x84, 0x41, 0x41, 0x73, 0x93, 0x33, 0xE4, 0x09, 0x84, 0x41,
        0x41, 0x73, 0x93, 0x33, 0xE4, 0x09, 0x84, 0x41, 0x41, 0x73, 0x93, 0x33, 0xE4, 0x09, 0x84, 0x41};
/**
r00 = 0x00000001     r01 = 0x00000074     r02 = 0x418409e4     r03 = 0x41840000     r04 = 0x00000001
r05 = 0x000000cd     r06 = 0x00000041     r07 = 0x00000009     r08 = 0x0000e409
r09 = 0x00004173     r10 = 0x0000dd73     r11 = 0x00000000     r12 = 0x00100020
r13 = 0x00100040     r14 = 0x00000000     r15 = 0x000001b6
**/
    if(argc == 1)
    {
        memcpy(&memory_bin[0x100020], testcase, 32);
        hexdump(&memory_bin[0x100000], 0x40);
        printf("starting\n");

        uint32_t res = do_stuff();
        printf("res=0x%x\n", res);
        hexdump(&memory_bin[0x100000], 0x40);
        assert(!memcmp(&memory_bin[0x100020], check, sizeof(check)));
        printf("test good\n");
    }
    else if (argc == 2 && strlen(argv[1]) == 64)
    {
        size_t sz = 64;
        unsigned char* in = hex2bin(argv[1], &sz);
        memcpy(&memory_bin[0x100020], in, 32);
        free(in);
        hexdump(&memory_bin[0x100000], 0x40);
        uint32_t res = do_stuff();
        printf("res=0x%x\n", res);
        hexdump(&memory_bin[0x100000], 0x40);
    }
    return 0;
}

uint32_t do_stuff()
{
    uint32_t regs[16]={0,};
#define r0 regs[0]
#define r1 regs[1]
#define r2 regs[2]
#define r3 regs[3]
#define r4 regs[4]
#define r5 regs[5]
#define r6 regs[6]
#define r7 regs[7]
#define r8 regs[8]
#define r9 regs[9]
#define r10 regs[10]
#define r11 regs[11]
#define r12 regs[12]
#define r13 regs[13]
#define r14 regs[14]
#define r15 regs[15]

"""
    for pc in range(0, (0x1b6+3), 3):
        ins = struct.unpack("<L", memory[pc:pc+4])[0] & 0xffffff
        asm, c = disassemble_ins(pc, ins)
        c_code += "    label_%04x:\n\n        %s\n\n" % (pc, c)
        print(asm)

    c_code += "}\n"
    sourcefile = "4_tz_vm_decompile.c"
    with open(sourcefile,"w") as f:
        f.write(c_code)
    if subprocess.call(["clang", "-Wno-unused-label", "-Ofast", "-flto", sourcefile]) == 0:
        subprocess.call(["./a.out"])

#----gdb---

SMC_0_ADDR = 0xe200c64
CALL_VM_RUN_INSTRUCTION = 0xE200CA4
TOS_FW_MEMCPY_ENTRY=0xE203484

"""
clang -arch arm64 -c 4_read_vm_memory_fast.S
segedit -extract __TEXT __text 4_read_vm_memory_fast.bin 4_read_vm_memory_fast.o

(__TEXT,__text) section
start:
0000000000000000	02 00 80 d2 	mov	x2, #0x0
0000000000000004	01 00 80 d2 	mov	x1, #0x0
loop:
0000000000000008	60 68 61 b8 	ldr		w0, [x3, x1]
000000000000000c	80 68 21 b8 	str		w0, [x4, x1]
0000000000000010	21 10 00 91 	add	x1, x1, #0x4
0000000000000014	3f 00 05 eb 	cmp		x1, x5
0000000000000018	81 ff ff 54 	b.ne	0x8
000000000000001c	00 00 20 d4 	brk	#0
"""
dump1000_shellcode = [
    0x02, 0x00, 0x80, 0xd2, 0x01, 0x00, 0x80, 0xd2, 0x60, 0x68, 0x61, 0xb8,
    0x80, 0x68, 0x21, 0xb8, 0x21, 0x10, 0x00, 0x91, 0x3f, 0x00, 0x05, 0xeb,
    0x81, 0xff, 0xff, 0x54, 0x00, 0x00, 0x20, 0xd4]

def gdb_dump_memory(addr, sz):
    inf = gdb.inferiors()[0]

    total=0x00101000
    CHUNK_SIZE=0x1000
    TOS_FW_MEMCPY_RET=TOS_FW_MEMCPY_ENTRY+len(dump1000_shellcode)-4

    inf.write_memory(TOS_FW_MEMCPY_ENTRY, bytearray(dump1000_shellcode))
    scratch_buffer = gdb.selected_frame().read_register("sp")-CHUNK_SIZE-0x1000
    data = b""
    bp = gdb.Breakpoint("* 0x%x" % (TOS_FW_MEMCPY_RET), internal=True)

    for off in range(0, sz, CHUNK_SIZE):
        gdb.execute("set $pc=0x%x" % TOS_FW_MEMCPY_ENTRY)
        gdb.execute("set $x3=0x%x" % (addr+off))
        gdb.execute("set $x4=0x%x" % scratch_buffer)
        gdb.execute("set $x5=0x%x" % CHUNK_SIZE)
        gdb.execute("c")
        data += inf.read_memory(scratch_buffer, CHUNK_SIZE)
    bp.delete()
    return data[:sz]

def save_ctx():
    frame =  gdb.selected_frame()
    regs = {}
    for r in ["pc", "sp"]:
        regs[r] = frame.read_register(r)
    for x in range(31):
        regs[r] = frame.read_register("x%d" % x)
    return regs

def restore_ctx(regs):
    frame =  gdb.selected_frame()
    regs = {}
    for r in list(regs.keys()):
        gdb.execute("set $%s=0x%x" % (r,regs[r]))

def gdb_dump_vm_registers():
    res = []
    regs = save_ctx()
    bp = gdb.Breakpoint("* 0x%x" % (SMC_0_ADDR+4), internal=True)
    bp.silent = True
    for reg in range(16):
        gdb.execute("set $pc=0x%x" % SMC_0_ADDR)
        gdb.execute("set $x3=0x0")
        gdb.execute("set $x2=0x0")
        gdb.execute("set $x1=0x%x" % reg)
        gdb.execute("set $x0=0x83010001")
        gdb.execute("c")
        value = gdb.selected_frame().read_register("x0")
        res.append(value)
    bp.delete()
    restore_ctx(regs)
    return res

def gdb_set_vm_register(reg, value):
    regs = save_ctx()
    bp = gdb.Breakpoint("* 0x%x" % (SMC_0_ADDR+4), internal=True)
    bp.silent = True
    gdb.execute("set $pc=0x%x" % SMC_0_ADDR)
    gdb.execute("set $x3=0x0")
    gdb.execute("set $x2=0x%x" % value)
    gdb.execute("set $x1=0x%x" % reg)
    gdb.execute("set $x0=0x83010002")
    gdb.execute("c")
    bp.delete()
    restore_ctx(regs)


def gdb_run_vm_instruction(ins):
    regs = save_ctx()
    bp = gdb.Breakpoint("* 0x%x" % (CALL_VM_RUN_INSTRUCTION+4), internal=True)
    bp.silent = True
    sp =  gdb.selected_frame().read_register("sp")
    gdb.execute("set $pc=0x%x" % CALL_VM_RUN_INSTRUCTION)
    gdb.execute("set $x1=0x%x" % (sp-4))
    gdb.execute("set $x0=0x%x" % ins)
    print("Running 0x%x" % ins)
    gdb.execute("c")
    return
    bp.delete()
    restore_ctx(regs)

def print_regs(regs):
    res = ""
    for i in range(len(regs)):
        res += "r%02d = 0x%08x    " % (i, regs[i])
        if i and (i%4) == 0:
            res += "\n"
    print(res)

def dump_memory_to_file(filename):
    data = gdb_dump_memory(0x0, 0x101000)
    print("Read 0x%x bytes to %s" % (len(data),filename))
    with open(filename, "wb") as f:
        f.write(data)

def main_gdb():
    if not gdb.inferiors()[0].was_attached:
        gdb.execute("set pagination off")
        gdb.execute("target remote :1234")
    pc = gdb.selected_frame().read_register("pc")
    if pc == 0:
        gdb.execute("break * 0xE2017B4")
        gdb.execute("c")
        print("Bypassing anti debug check")
        gdb.execute("set $x0=0x0")
        gdb.execute("break * 0xE200C80")
        gdb.execute("c")
        dump_memory_to_file(MEMORY_DUMP_FILE)
        gdb.execute("kill")
        exit(0)
    print_regs(gdb_dump_vm_registers())
    dump_memory_to_file(MEMORY_DUMP_FILE)
    return
    memory = open(MEMORY_DUMP_FILE, "rb").read()
    inf = gdb.inferiors()[0]

    run_prog = True
    run_prog = False
    #run_prog = True
    instructions = [0x12c000]

    if run_prog:
        for i in range(16):
            gdb_set_vm_register(i, 0)
    else:
        for i in range(15):
            gdb_set_vm_register(i, int("0x%x%x%x%x%x%x%x%x" % (i,i,i,i,i,i,i,i),16))
        gdb_set_vm_register(15,0)
        gdb_set_vm_register(0,0xaabbccdd) 
        gdb_set_vm_register(11,0x4242)
    init_regs = gdb_dump_vm_registers()
    last_pc = init_regs[-1]

    for _ in range(500):
        if run_prog:
            ins = struct.unpack("<L", memory[last_pc:last_pc+4])[0] & 0xffffff
        else:
            ins = instructions.pop()
        disas,_ = disassemble_ins(last_pc, ins)
        print(disas)
        gdb_run_vm_instruction(ins)
        new_regs = gdb_dump_vm_registers()
        if init_regs[:-1] != new_regs[:-1]:
            print("change after 0x%x" % ins)
            print_regs(new_regs)
        last_pc = new_regs[-1]
        old_regs = new_regs

def main():
    if "gdb" in globals():
        return main_gdb()
    return main_generate_c()

if __name__ == "__main__":
    main()
