#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <assert.h>
#include "4_vm_memory.h"

typedef union
{
    uint8_t tables[256][256][256];
    uint8_t u8[0x1000000];
} tables_t;

typedef union
{
    uint64_t u64[4];
    uint32_t u32[8];
    uint16_t u16[16];
    uint8_t u8[32];
} u256_t;

typedef union
{
    uint64_t u64;
    uint32_t u32[2];
    uint16_t u16[4];
    uint8_t u8[8];
} u64_t;

#define _byteswap_ushort(x) __builtin_bswap16(x)

unsigned char* hex2bin(const char* hexstr, size_t* size)
{
    size_t hexstrLen = strlen(hexstr);
    size_t bytesLen = hexstrLen / 2;
    
    unsigned char* bytes = (unsigned char*) malloc(bytesLen);
    
    int count = 0;
    const char* pos = hexstr;
    
    for(count = 0; count < bytesLen; count++) {
        sscanf(pos, "%2hhx", &bytes[count]);
        pos += 2;
    }
    
    if( size != NULL )
        *size = bytesLen;
    
    return bytes;
}

static inline uint8_t BYTE1(uint32_t x)
{
    return (x>>8) & 0xff;
}
static inline uint8_t BYTE0(uint32_t x)
{
    return x & 0xff;
}

tables_t* constants = (tables_t*) &__4_vm_memory_bin[0x1000];

const int tables_idx[32*4] = {7, 6, 5, 4, 3, 2, 1, 0, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0};

uint8_t invert_lookup(uint8_t x, uint8_t y, uint8_t z)
{
    for(int i=0; i < 256; i++)
    {
        if(constants->tables[tables_idx[x]][i][y] == z)
        {
            return i;
        }
    }
    assert("fail");
    return 0;
}

void transform(u64_t* state, int j)
{
    uint8_t a,b;
    int dostuff = (((uint8_t)j - 1) & 8) == 0;
    uint8_t x = 128-j*4;

    uint16_t tmp = state->u16[2];
    state->u16[2] = state->u16[3];
    state->u16[3] = state->u16[0];

    a = constants->tables[tables_idx[x]][BYTE0(state->u16[1])][BYTE1(state->u16[1])];

    assert(invert_lookup(x, BYTE1(state->u16[1]), a) == BYTE0(state->u16[1]));
    
    b = constants->tables[tables_idx[x+1]][BYTE1(state->u16[1])][a];

    assert(invert_lookup(x+1, a, b) == BYTE1(state->u16[1]));

    state->u8[0] = constants->tables[tables_idx[x+2]][a][b];
    assert(invert_lookup(x+2, b, state->u8[0]) == a);

    state->u8[1] = constants->tables[tables_idx[x+3]][b][state->u8[0]];
    
    assert(invert_lookup(x+3, state->u8[0], state->u8[1]) == b);
    
    if (dostuff)
        state->u16[3] ^= (j ^ state->u16[1]);
    state->u16[1] = tmp;
    if(!dostuff)
        state->u16[1] ^= (j ^ state->u16[0]);
}

void invert_transform(u64_t* state, uint8_t j)
{
    uint8_t a,b;
    uint16_t tmp;
    int dostuff = (((uint8_t)j - 1) & 8) == 0;
    uint8_t x = 128-j*4;

    if(!dostuff)
        state->u16[1] ^= (j ^ state->u16[0]);
    tmp = state->u16[1];

    b = invert_lookup(x+3, state->u8[0], state->u8[1]);
    a = invert_lookup(x+2, b, state->u8[0]);
    state->u8[2+1] = invert_lookup(x+1, a, b);
    state->u8[2] = invert_lookup(x, BYTE1(state->u16[1]), a);

    if (dostuff)
        state->u16[3] ^= (j ^ state->u16[1]);

    state->u16[0] = state->u16[3];
    state->u16[3] = state->u16[2];
    state->u16[2] = tmp;
}

uint32_t do_stuff(u256_t* input)
{
    u64_t state;
    u64_t bak;
    u64_t check;

    for(int i=0; i < 4; i++)
    {
        state.u16[0] = _byteswap_ushort(input->u16[i*4]);
        state.u16[1] = _byteswap_ushort(input->u16[i*4+1]);
        state.u16[2] = _byteswap_ushort(input->u16[i*4+2]);
        state.u16[3] = _byteswap_ushort(input->u16[i*4+3]);
        for(int j=32; j > 0; j--)
        {
            bak = state;
            transform(&state, j);
            check = state;
            invert_transform(&check, j);
            assert(!memcmp(&check, &bak, sizeof(check)));
        }
        memcpy(&input->u16[i*4], &state, 8);

    }
    for(int i=0; i < 16; i++)
    {
        input->u16[i] = _byteswap_ushort(input->u16[i]);
    }
    return 0;
}

void reverse_and_print(u256_t* input)
{
    u64_t state;

    for(int i=0; i < 16; i++)
    {
        input->u16[i] = _byteswap_ushort(input->u16[i]);
    }
    
    for(int i=0; i < 4; i++)
    {
        memcpy(&state, &input->u16[i*4], 8);

        for(int j=1; j <= 32; j++)
        {
            invert_transform(&state, j);
        }
        input->u16[i*4] = _byteswap_ushort(state.u16[0]);
        input->u16[i*4+1] = _byteswap_ushort(state.u16[1]);
        input->u16[i*4+2] = _byteswap_ushort(state.u16[2]);
        input->u16[i*4+3] = _byteswap_ushort(state.u16[3]);
        
    }
    for(int i=0; i < 32; i++)
    {
        printf("%02x", input->u8[i]);
    }
    printf("\n");
}

int main(int argc, const char * argv[]) {
    uint8_t testcase[32] = {0};
    memset(testcase, 0xAA, 32);
    const uint8_t check[32] = {
        0x41, 0x73, 0x93, 0x33, 0xE4, 0x09, 0x84, 0x41, 0x41, 0x73, 0x93, 0x33, 0xE4, 0x09, 0x84, 0x41,
        0x41, 0x73, 0x93, 0x33, 0xE4, 0x09, 0x84, 0x41, 0x41, 0x73, 0x93, 0x33, 0xE4, 0x09, 0x84, 0x41};

    if(argc == 1)
    {
        do_stuff((u256_t*)testcase);
        assert(!memcmp(&testcase, check, sizeof(check)));
        printf("testcase OK\n");
    }
    else if (argc == 2 && strlen(argv[1]) == 32)
    {
        reverse_and_print((u256_t*) argv[1]);
    }
}
