# SSTIC 2019 #

Solution minimaliste du challenge SSTIC 2019.

    docker build --tag=sstic2019 .
    docker run -t -v $(pwd):/sstic sstic2019 make -C /sstic

## Introduction ##

Le challenge du SSTIC 2019 se présente sous la forme d'un système Linux ARM64 avec un firmware dérivé de l'implémentation de référence Trusted Firmware-A d'ARM (https://github.com/ARM-software/arm-trusted-firmware).

L'outil fiptool du projet arm-trusted-firmware permet d'identifier les différents composants.

    ./fiptool info virtual_phone/flash.bin
    Trusted Boot Firmware BL2: offset=0xD8, size=0x9430, cmdline="--tb-fw"
    EL3 Runtime Firmware BL31: offset=0x9508, size=0x90A0, cmdline="--soc-fw"
    Secure Payload BL32 (Trusted OS): offset=0x125A8, size=0x5370, cmdline="--tos-fw"
    Non-Trusted Firmware BL33: offset=0x17918, size=0x2220020, cmdline="--nt-fw"

Initialement, tous les composants du fichier flash.bin sont chiffrés.

Le firmware ARM a été modifié pour manipuler un keystore (fichier sur le système hôte exécutant QEMU), via la fonctionnalité de semihosting de QEMU. Le keystore sert à stocker les clés de déchiffrement des différentes étapes du challenge. 

## Etape 1 ##

La première étape consiste à retrouver une clé privée RSA permettant de déchiffrer une clé AES dans le keystore initial, afin que le code de la rom (rom.bin) puisse charger les composants du firmware nécessaires au démarrage du système.

En visualisant avec matplotlib la trace de consommation de courant fournie (power_consumtion.npz), on distingue une zone qui semble permettre d'identifier les bits de la clé via une SPA (Simple Power Analysis, https://en.wikipedia.org/wiki/Power_analysis#/media/File:Power_attack_full.png). 

De manière empirique, une succession de plus de 4 "pics" correspond à un 1.
Le script 1_power_analysis.py retrouve l'exposant privé :

    0x23d87cdf97bb95abe6273c384190c765f552ab86f6de30a8db74435c95e6e3138f54af689812d8f9359cf0f4d453a0c11ec68ce470216c09e74c8947adaf23e902415d61ddf2c0ffe459cbb40f7de42bdb7cd14093100a570e8c29819765e2d8d276f86471b52ac29aa2ce2bb72cd45006279e82bec253ae9675fe45824f6001

Ce qui permet de déchiffrer la clé AES :

    SSTIC{a947d6980ccf7b87cb8d7c246}

Cette clé permet de démarrer le système Linux et de déchiffrer les composants du firmware :

* TRUSTED_BOOT_FIRMWARE_BL2 : bootloader intermédiaire (37 Ko)
* EL3_RUNTIME_FIRMWARE_BL31 : Secure Monitor (36 Ko)
* NON_TRUSTED_FIRMWARE_BL33 : firmware UEFI, noyau Linux et ramdisk (34 Mo)

Le Secure Monitor est le code le plus privilégié du système (EL3 Secure World), il gère l'interfaçage avec le keystore via le semihosting, ainsi que des fonctionnalités utilisée dans l'étape 4.
Au niveau Linux, un module noyau sstic.ko est présent et expose un périphérique /dev/sstic, qui fait l'interface entre le Secure Monitor et l'userland. Cette interface est utilisée par le script /root/tools/keystore.py pour lire et écrire les clés des différentes étapes dans le keystore, ainsi que par le binaire de l'étape 4 (/root/safe_02/decrypted_file).

## Etape 2 ##

L'étape 2 est un script Python (get_safe1_key.py) a compléter à l'aide du schéma de circuit logique d'un "Secure Element" (schematics.png). Les informations suivantes sont fournies dans le script :

    """
    TODO : 
        - Implémentation de la communication avec le secure element
            * Entrées A et B (A0 = bit de poids faible)
            * Entrée OP
            * Sortie Out (Out0 = bit de poids faible)
            * Les boutons permettent à l'utilisateur de rentrer sa combinaison secrète pour le déchiffrement
        - Supprimer les docs de conception 
    """

Le schéma est composé de nombreux multiplexeurs 2-vers-1 qui construisent les fonctions suivantes :

* rotation à gauche de 1 bit des entrés A et B activées par les boutons 3 et 4
* ET/OU/XOR/ADD des entrées A et B (après rotation éventuelle), l'opération est choisie selon l'état des boutons 1 et 2 et l'entrée OP

Ce qui permet de compléter le script Python :

    def secure_device(a,b,op):
        if B & 0b0100:
            a = rol1(a)
        if B & 0b1000:
            b = rol1(b)

        sel = op ^ (B & 0b11)
        if sel == 0:
            return a&b
        elif sel == 1:
            return a|b
        elif sel == 2:
            return a^b
        elif sel == 3:
            return (a+b) & 0xff

Le script attend ensuite 8 combinaisons des 4 boutons et compare le SHA256 de la sortie à une valeur fixe.

Un bruteforce en C des 2^32 combinaisons possibles a été écrit en utilisant Sharlotte256 (https://github.com/cashaddress/Sharlotte256). La clé est trouvée en environ 10 minutes.

    ./2_bruteforce_safe1_key
    x=0x646b235e
    key= 8FA4DFA9D4EDBBF0

    python3 ./2_get_safe1_key.py
    [i] Dechiffrement du conteneur

    [i] Initialisation du secure element
    [i] Merci d'appuyer les 4 boutons du secure element puis appuyer sur entrée
       B = 0b1111
    [+] Initialisation, check 1 OK
    [i] Merci de relâcher les 4 boutons du secure element puis appuyer sur entrée
       B = 0b0000
    [+] Initialisation, check 2 OK
    [w] Calcul de la clef de déchiffrement du conteneur, préparez vous à appuyer sur les boutons
    [i] Step 1, appuyez sur entrée quand la combinaison de boutons est enfoncée
       B = 0b1110
    [i] Step 2, appuyez sur entrée quand la combinaison de boutons est enfoncée
       B = 0b0101
    [i] Step 3, appuyez sur entrée quand la combinaison de boutons est enfoncée
       B = 0b0011
    [i] Step 4, appuyez sur entrée quand la combinaison de boutons est enfoncée
       B = 0b0010
    [i] Step 5, appuyez sur entrée quand la combinaison de boutons est enfoncée
       B = 0b1011
    [i] Step 6, appuyez sur entrée quand la combinaison de boutons est enfoncée
       B = 0b0110
    [i] Step 7, appuyez sur entrée quand la combinaison de boutons est enfoncée
       B = 0b0100
    [i] Step 8, appuyez sur entrée quand la combinaison de boutons est enfoncée
       B = 0b0110
    [+] Hash ok
    [i] Dérivation de la clef AES safe_01
    [i] aes key : 5fb3a83d1fd97137076019ad6e96c6a366fb6b32618d162e00cdee9bad427a8a
    [i] Vous pouvez sauvegarder cette clef en utilisant /root/tools/add_key.py key

## Etape 3 ##

La troisième étape est un binaire statique presque vide (à part la libc statique) qui semble faire une utilisation détournée des exceptions C++.
En plaçant un watchpoint en lecture sur argv[1], on tombe dans une fonction de libgcc.so.

Il s'agit de la fonction execute_stack_op (https://github.com/gcc-mirror/gcc/blob/master/libgcc/unwind-dw2.c#L516). La vérification du flag est en fait implémentée en bytecode DWARF (cf 
"Exploiting the Hard-Working DWARF" https://www.usenix.org/legacy/events/woot11/tech/slides/oakley.pdf). La VM DWARF est une machine à pile.

Le script 3_dwarf_vm.py désassemble le bytecode DWARF et génère un code C équivalent pour chaque instruction (3_dwarf_decompile.c). Ce code C est ensuite compilé en -O3, et la sortie de la décompilation Hex-Rays de ce programme a ensuite été simplifiée manuellement, ce qui a pris un certain temps...

Le programme 3_dwarf_reverse.c contient le code simplifié, ainsi que le code pour inverser la transformation de l'entrée utilisateur et retrouver le flag à partir des valeurs de sortie attendues pour afficher le message "That's the correct flag :)". L'algorithme implémenté par le bytecode n'a pas été identifié.

    ./3_dwarf_reverse
    testcase ok
    53 53 54 49 43 7B 44 77  34 72 66 5F 56 4D 5F 31  |  SSTIC{Dw4rf_VM_1
    73 5F 63 6F 30 6C 5F 69  73 6E 5F 74 5F 49 74 7D  |  s_co0l_isn_t_It}

Cette clé permet de déchiffrer le dernier binaire, ainsi que le composant SECURE_PAYLOAD_BL32 du firmware (alias Trusted OS, alias Secure EL1). Un redémarrage est nécessaire pour charger le Trusted OS et pouvoir utiliser le binaire de l'étape 4.

## Etape 4 ##

Le binaire /root/safe_03/decrypted_file utilise le périphérique /dev/sstic pour charger le bytecode (chiffré) de vérification du flag dans une machine virtuelle (à registres) implémentée dans le Secure World.

La machine virtuelle est implémentée dans les 2 parties du Secure World:

* BL32 : lecture et décodage des instructions, gestion de la mémoire (chiffrée) de la VM via un handler d’exception. Implémentation d'une partie des instructions, dont certaines dans du code ARM 32 bits.
* BL31 : gestion des registres de la VM (chiffrés), implémentations d'une autre partie des instructions.

Comme ces 2 parties utilisent des instructions ARMv8.2 non supportés par IDA (par exemple SM4EKEY), Ghidra a été utilisé. Le patch suivant permet de suivre les entrées/sorties de l'instruction SMC utilisée pour communiquer entre BL32 et BL31.

    --- Ghidra/Processors/AARCH64/data/languages/AARCH64base.sinc.bak
    +++ Ghidra/Processors/AARCH64/data/languages/AARCH64base.sinc
    @@ -5196,7 +5196,7 @@
    :smc imm16
    is b_2431=0xd4 & excCode=0 & imm16 & excCode2=0 & ll=3
    {
    -	CallSecureMonitor(imm16:2);
    +	x0 = CallSecureMonitor(imm16:2, x0, x1, x2, x3);
    }

    # C6.2.217 SMNEGL page C6-865 line 50269 MATCH x9b20fc00/mask=xffe0fc00

Un anti-debug est présent dans BL32, la fonction 0xE201650 désobfusque des chaînes de caractères et utilise le semihosting pour lire le fichier /proc/self/cmdline ou executer "(ps -ef |egrep -q 'qemu-system.* -[s] |qemu-system.* -[g]db ') && exit 42" sur l'hôte.

L'interface Python de gdb a été utilisée pour :

* Déchiffrer la mémoire de la VM au début de l’exécution du programme via l'injection d'un shellcode qui recopie la zone mémoire 0x0-0x101000 (en déclenchant le handler d'exception à chaque lecture d'un mot de 32 bits)
* Lire et écrire les registres de la VM via les appels SMC 0x83010001 et 0x83010002
* Exécuter une instruction arbitraire après avoir initialisé les registres et observer leur état de sortie pour en déduire le type d'instruction

Le script 4_tz_vm.py contient le désassembleur et les scripts gdb. Comme pour l'étape 3, le désassembleur génère du code C qui a été compilé puis décompilé, simplifié et inversé (4_tz_vm_reverse.c). L'algorithme implémenté par le bytecode n'a pas été identifié.

La clé est obtenue en inversant l'état final de l'algorithme correspondant à la chaine "pr.a.rfg.cnf.fv.snpvyr@ffgvp.bet" (ce.n.est.pas.si.facile@sstic.org en rot13).

    ./4_tz_vm_reverse "pr.a.rfg.cnf.fv.snpvyr@ffgvp.bet"
    acadaa8b5b55306fb3c6dfc3b2d1c80770084644225febd71a9189aa26ec740e

    python3 ./decrypt_container.py root/safe_03 acadaa8b5b55306fb3c6dfc3b2d1c80770084644225febd71a9189aa26ec740e

    tar xf root/safe_03/decrypted_file data/com.android.providers.telephony/databases/
    sqlite3 data/com.android.providers.telephony/databases/mmssms.db "select body from sms" | grep cactus
    Mission accomplie, comme d'habitude la perspective d'une tournée de shooter au cactus a suffi à corrompre le CO et à chosir la date de publication du challenge. Pour être sûr que les experts sont toujours occupés, j'ai redirigé l'adresse 9e915a63d3c4d57eb3da968570d69e95@challenge.sstic.org vers votre boîte mail. Tant que vous ne voyez passer aucun mail, la voie est libre...

Merci aux organisateurs !
