#!/usr/bin/env python3

import argparse
import struct
from Crypto.Cipher import AES

KEY1 = "SSTIC{a947d6980ccf7b87cb8d7c246}"
KEY_BL32 = "SSTIC{Dw4rf_VM_1s_co0l_isn_t_It}"

#from arm-trusted-firmware/include/tools_share/firmware_image_package.h 
TOC_HEADER_NAME = 0xAA640001

UUID_TRUSTED_BOOT_FIRMWARE_BL2 = \
    [0x5f, 0xf9, 0xec, 0x0b, 0x4d, 0x22, 0x3e, 0x4d, 0xa5, 0x44, 0xc3, 0x9d, 0x81, 0xc7, 0x3f, 0x0a]
UUID_EL3_RUNTIME_FIRMWARE_BL31 = \
    [0x47,  0xd4, 0x08, 0x6d, 0x4c, 0xfe, 0x98, 0x46, 0x9b, 0x95, 0x29, 0x50, 0xcb, 0xbd, 0x5a, 0x00]
UUID_SECURE_PAYLOAD_BL32 = \
    [0x05,  0xd0, 0xe1, 0x89, 0x53, 0xdc, 0x13, 0x47, 0x8d, 0x2b, 0x50, 0x0a, 0x4b, 0x7a, 0x3e, 0x38]
UUID_NON_TRUSTED_FIRMWARE_BL33 = \
    [0xd6,  0xd0, 0xee, 0xa7, 0xfc, 0xea, 0xd5, 0x4b, 0x97, 0x82, 0x99, 0x34, 0xf2, 0x34, 0xb6, 0xe4]

def parse_and_decrypt_fip(filename, key):
    assert key in [KEY1, KEY_BL32]

    f = open(filename, "rb")
    name, serial_number, flags = struct.unpack("<LLQ", f.read(16))
    assert name == TOC_HEADER_NAME
    g = globals()
    entries = []
    while True:
        uuid = list(f.read(16))
        if sum(uuid) == 0:
            break
        offset_address, size, flags = struct.unpack("<QQQ", f.read(24))
        entries.append((uuid, offset_address, size, flags))
    
    for uuid, offset_address, size, flags in entries:
        name = None
        for k in g:
            if k.startswith("UUID_") and g[k] == uuid:
                name = k.replace("UUID_", "") + ".bin"
                break
        assert name
        if key == KEY1:
            if uuid == UUID_SECURE_PAYLOAD_BL32:
                continue
        elif uuid != UUID_SECURE_PAYLOAD_BL32:
            continue
        f.seek(offset_address)
        data = f.read(size)
        data = decrypt_image(data, key)
        print(("Writing %s" % name))
        with open(name, "wb") as f2:
            f2.write(data)
        if uuid == UUID_NON_TRUSTED_FIRMWARE_BL33:
            print("Writing initrd.cpio (from NON_TRUSTED_FIRMWARE_BL33)")
            with open("initrd.cpio", "wb") as f2:
                f2.write(data[0xC20000:])

def decrypt_image(data, key):
    return AES.new(key, AES.MODE_CBC, data[:16]).decrypt(data[16:])

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("flashbin")
    parser.add_argument("key")
    args = parser.parse_args()
    parse_and_decrypt_fip(args.flashbin, args.key)

if __name__ == "__main__":
    main()
